# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cloudpickle',
 'common_client_scheduler',
 'terality',
 'terality._terality',
 'terality._terality.bin',
 'terality._terality.data_transfers',
 'terality._terality.encoding',
 'terality._terality.terality_structures',
 'terality._terality.utils',
 'terality_serde']

package_data = \
{'': ['*']}

install_requires = \
['aiobotocore>=1.4.2,<2.0.0',
 'boto3>=1.17.106,<1.17.107',
 'click>=8.0.1,<9.0.0',
 'numpy>=1.18,<2.0',
 'pandas>=1.0.0,<2.0.0',
 'pyarrow>=6.0.0,<7.0.0',
 'pydantic>=1.7.3,<2.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'requests>=2.26.0,<3.0.0',
 'sentry-sdk>=1.3.0,<2.0.0',
 'single-source>=0.1.5,<0.2.0',
 'typing_extensions>=3.10.0,<4.0.0',
 'urllib3>=1.26.0']

extras_require = \
{'azure': ['azure-storage-file-datalake>=12.4.0,<13.0.0',
           'azure-storage-blob>=12.8.1,<13.0.0',
           'azure-identity>=1.6.0,<2.0.0']}

entry_points = \
{'console_scripts': ['terality = terality._terality.bin.__main__:cli']}

setup_kwargs = {
    'name': 'terality',
    'version': '0.12.17',
    'description': 'The Data Processing Engine for Data Scientists',
    'long_description': 'Terality is the serverless & lightning fast Python data processing engine.\n\nTerality’s engine enable data scientists and engineers to transform and manipulate data at light speed, with the exact same syntax as Pandas, with zero sever management.\n\nYou will need a Terality account to use this package. Check out our documentation to get started https://docs.terality.com/.',
    'author': 'Terality Engineering Team',
    'author_email': 'dev.null@terality.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://terality.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



from io import BytesIO
from typing import Any, Callable, Hashable, List, Optional, Tuple

import numpy as np
import pandas as pd
import pyarrow.lib
import terality

from common_client_scheduler.structs import (
    IndexType,
    ParquetMetadataIndex,
    ParquetMetadataCols,
    ParquetMetadata,
)
from terality.exceptions import TeralityClientError

from terality_serde import CallableWrapper
from common_client_scheduler import (
    StructRef,
    PandasIndexMetadata,
    PandasSeriesMetadata,
    PandasDFMetadata,
    NDArrayMetadata,
)
from common_client_scheduler.requests_responses import AwsCredentials
from terality_serde.json_encoding import (
    ArrayEncoding,
    _endecode_json_indices_inplace,
    _endecode_json_columns_inplace,
    has_float_nan,
    has_only_jsonable_types,
)

from .. import DataTransfer, global_client


def _check_if_array_needs_encoding(
    array: pd.api.extensions.ExtensionArray,
) -> Optional[ArrayEncoding]:
    if isinstance(array, pd.arrays.TimedeltaArray):
        return ArrayEncoding.DURATION

    if array.dtype.name != "object":
        return None

    ndarray = array.to_numpy()

    inferred = pd.api.types.infer_dtype(ndarray)
    if inferred == "string":
        if has_float_nan(ndarray):
            return ArrayEncoding.MIXED
        return None

    if has_only_jsonable_types(ndarray):
        return ArrayEncoding.MIXED

    return ArrayEncoding.STRUCT


def _extract_metadata(df: pd.DataFrame) -> ParquetMetadata:
    if isinstance(df.index, pd.DatetimeIndex) and df.index.freq is not None:
        datetime_index_freq = df.index.freq.name
    else:
        datetime_index_freq = None

    indices_encoding = [
        _check_if_array_needs_encoding(df.index.get_level_values(num).array)
        for num in range(df.index.nlevels)
    ]
    index_metadata = ParquetMetadataIndex(
        type_=_pandas_index_type_to_enum(df.index),
        encodings=indices_encoding,
        names=list(df.index.names),
        name=df.index.name,
        datetime_index_freq=datetime_index_freq,
    )

    metadata_cols = ParquetMetadataCols(
        encodings=[_check_if_array_needs_encoding(df[col].array) for col in df],
        col_names=list(df.columns),
        col_names_name=df.columns.name,
    )

    metadata = ParquetMetadata(index=index_metadata, cols=metadata_cols)
    return metadata


def _encode_inplace_and_upload_df(
    aws_credentials: AwsCredentials, df: pd.DataFrame
) -> Tuple[str, ParquetMetadata]:
    """Parquet only supports dataframes so we transmit all pandas structures as dataframes"""

    # Get metadata which would be lost through parquet
    metadata = _extract_metadata(df)

    # Make df parquet compatible
    # WARNING: Parquet only accepts string column names
    # => We replace non-string column names with placeholders and pass the columns "out of band" (in the metadata).
    df = df.rename(
        columns={df.columns[col_num]: str(col_num) for col_num in range(len(df.columns))},
        copy=False,
    )
    _endecode_json_indices_inplace(df, metadata.index.encodings, encode=True)
    _endecode_json_columns_inplace(df, metadata.cols.encodings, encode=True)

    # Upload df
    row_group_size = _estimate_row_group_size(df, 1024 ** 3)
    data_bytes = BytesIO()
    df.to_parquet(data_bytes, version="2.6", row_group_size=row_group_size)
    import_id = DataTransfer.upload_bytes(
        global_client().get_upload_config(), aws_credentials, data_bytes  # type: ignore
    )

    return import_id, metadata


def _estimate_row_group_size(df: pd.DataFrame, max_memory_per_row_group_bytes=1024 ** 3) -> int:
    """Estimate the size of a parquet row group that can be processed in memory for this dataframe,
    index included.

    In other words, split the dataframe in row groups, each row group not exceeding
    `max_memory_per_row_group_bytes`, and return the size (in rows) of a row group.

    Server-side, we can only process parquet files by row group. We can not split a parquet file in the
    middle of a row group. This means that a whole uncompressed row group must fit in memory. In order to
    maximize parallelisation server-side, this function returns how many rows per row group the parquet file
    should contain.

    We still want this value to be as high as possible to improve compression and reduce the bandwidth
    used, but err on the side of caution - an incorrect split means we can't easily process the uploaded
    files (we risk returning memory errors, which would be unfortunate for Terality :) ).

    We also want this operation to be as fast as possible, so we don't use `memory_usage(deep=True)`, which
    is costly on big datasets. Instead, for columns with dtype = object, we take a guess based on a small
    sample and hope it covers all real-world usage. This function may fail on dataframes that have a wide
    distribution of object sizes.
    """
    # as a safety measure, never include more than this many rows per row group.
    # this should give us a safety net even if our memory usage estimation is wildly inaccurate.
    MAX_ROWS_PER_ROW_GROUP = 2_000_000  # pylint: disable=invalid-name

    memory_usage_bytes = _estimate_dataframe_memory_usage_bytes(df)
    rows_per_row_group = (len(df) * max_memory_per_row_group_bytes) // memory_usage_bytes

    return min(rows_per_row_group, MAX_ROWS_PER_ROW_GROUP)


def _estimate_dataframe_memory_usage_bytes(
    df: pd.DataFrame, max_samples: int = 500, random_state=None
) -> int:
    """Estimate the memory usage of a DataFrame, in bytes.

    See _estimate_column_memory_usage_bytes for the difference with `memory_usage(deep=True)`.
    """
    # Get memory usage for all columns, except when dtype = object.
    memory_usage = df.memory_usage(deep=False)
    total_memory_bytes = 0

    object_columns = [column for column, dtype in df.dtypes.iteritems() if dtype == "object"]
    nonobject_columns = [column for column, dtype in df.dtypes.iteritems() if dtype != "object"]

    for nonobject_column in nonobject_columns:
        total_memory_bytes += memory_usage[nonobject_column]

    estimate_index_memory_use = True
    # Index could be a MultiIndex too, without a single dtype. For those, we'll also rely on sampling.
    if isinstance(df.index, pd.Index) and df.index.dtype != "object":
        # Use exact memory usage for single-level indexes with a non-object dtype.
        total_memory_bytes += memory_usage["Index"]
        estimate_index_memory_use = False

    total_memory_bytes += _estimate_columns_memory_usage_bytes(
        df, object_columns, estimate_index_memory_use, max_samples, random_state
    )

    return total_memory_bytes


def _estimate_columns_memory_usage_bytes(
    df: pd.DataFrame,
    columns: List[Hashable],
    include_index: bool,
    max_samples: int = 200,
    random_state=None,
) -> int:
    """Estimate the memory usage of columns in a DataFrame, in bytes.

    Unlike memory_usage(deep=True), this function samples rows and extrapolate the memory usage from these
    rows only. This makes it faster, but inaccurate.

    If `columns` is an empty list, the memory usage of the index will still be included in the return when
    include_index is True.

    The sample is random, and thus the results of this function are not reproducible unless random_state is
    provided.
    """
    df_columns: pd.DataFrame = df[columns]
    samples_count: int = min(max_samples, len(df_columns))
    samples: pd.DataFrame = df_columns.sample(n=samples_count, random_state=random_state)
    return (
        samples.memory_usage(deep=True, index=include_index).sum().squeeze()
        / samples_count
        * len(df_columns)
    )


def _pandas_index_type_to_enum(index: pd.Index) -> IndexType:
    """
    NOTE: DatetimeIndex inherit Int64Index, so keep following assertion order.
    """

    if isinstance(index, pd.MultiIndex):
        return IndexType.MULTI_INDEX
    if isinstance(index, pd.DatetimeIndex):
        return IndexType.DATETIME_INDEX
    if isinstance(index, pd.Int64Index):
        return IndexType.INT64_INDEX
    if isinstance(index, pd.Float64Index):
        return IndexType.FLOAT64_INDEX
    return IndexType.INDEX


# TODO the above version is more opti to upload/store data,
#  check how to read with numpy scheduler side (ndarray_from_numpy_metadata)
def _upload_ndarray_and_get_metadata(
    aws_credentials: AwsCredentials, ndarray: np.ndarray
) -> NDArrayMetadata:
    df = pd.DataFrame(ndarray)
    transfer_id, metadata = _encode_inplace_and_upload_df(aws_credentials, df)
    return NDArrayMetadata(transfer_id, metadata)


def _upload_index_and_get_metadata(
    aws_credentials: AwsCredentials, index: pd.Index
) -> PandasIndexMetadata:
    df = index.to_frame(index=True).iloc[:, 0:0]
    transfer_id, metadata = _encode_inplace_and_upload_df(aws_credentials, df)
    return PandasIndexMetadata(transfer_id, metadata)


def _upload_series_and_get_metadata(
    aws_credentials: AwsCredentials, series: pd.Series
) -> PandasSeriesMetadata:
    # not series.to_frame() to preserve name when name is None, which will get replaced by 0
    df = pd.DataFrame({series.name: series.copy(deep=True)})
    transfer_id, metadata = _encode_inplace_and_upload_df(aws_credentials, df)
    return PandasSeriesMetadata(transfer_id, metadata)


def _upload_df_and_get_metadata(
    aws_credentials: AwsCredentials, df: pd.DataFrame
) -> PandasDFMetadata:
    df_for_upload = df.copy(deep=True)
    transfer_id, metadata = _encode_inplace_and_upload_df(aws_credentials, df_for_upload)
    return PandasDFMetadata(transfer_id, metadata)


def _to_terality(obj: Any) -> Any:
    if isinstance(obj, np.ndarray):
        return terality.NDArray.from_numpy(obj)
    if isinstance(obj, pd.Index):
        return terality.Index.from_pandas(obj)
    if isinstance(obj, pd.Series):
        return terality.Series.from_pandas(obj)
    if isinstance(obj, pd.DataFrame):
        return terality.DataFrame.from_pandas(obj)
    return obj


def _get_upload_method(obj: Any) -> Optional[Callable[[AwsCredentials, Any], Any]]:
    if isinstance(obj, np.ndarray):
        if len(obj.shape) != 1:
            raise TeralityClientError("Terality only supports 1D numpy array.")
        return _upload_ndarray_and_get_metadata
    if isinstance(obj, pd.Index):
        return _upload_index_and_get_metadata
    if isinstance(obj, pd.Series):
        return _upload_series_and_get_metadata
    if isinstance(obj, pd.DataFrame):
        return _upload_df_and_get_metadata
    return None


def _upload_object_and_get_metadata(
    credentials_fetcher: AwsCredentials, upload_method: Callable, obj: Any
):
    """
    When XXX.from_pandas is called, we upload pandas object data in S3 and we send metadata to the scheduler.
    The scheduler will then deserialize the metadata, then build and return a Terality structure to the client.
    """

    try:
        return upload_method(credentials_fetcher, obj)
    except pyarrow.lib.ArrowNotImplementedError as e:
        message = str(e)
        # Simplify the user facing error message
        message = message.replace("Unhandled type for Arrow to Parquet schema conversion: ", "")
        raise TeralityClientError(
            f"Terality can not transfer the data structure: unsupported data type: {message}"
        ) from e


def encode(credentials_fetcher: AwsCredentials, function_name: str, obj: Any):
    from terality import NDArray, Index, Series, DataFrame  # Avoid circular dependencies
    from terality._terality.terality_structures import DataFrameGroupBy, SeriesGroupBy

    if function_name in ["from_pandas", "from_numpy"]:
        upload_method = _get_upload_method(obj)
        if upload_method is not None:
            return _upload_object_and_get_metadata(credentials_fetcher, upload_method, obj)
        raise ValueError(f"Can't convert type {type(obj)} to terality structure")
    try:
        obj = _to_terality(obj)
    except Exception as e:
        raise ValueError(
            f"An error occured while converting parameter {obj} into a Terality structure"
        ) from e

    if isinstance(obj, (NDArray, Index, Series, DataFrame, DataFrameGroupBy, SeriesGroupBy)):
        # noinspection PyProtectedMember
        return StructRef(id=obj._id, type=obj.__class__._class_name)
    if callable(obj) and not isinstance(obj, type):  # avoid types, which are also callables
        return CallableWrapper.from_object(obj)

    return obj



from .utils import (
    logger,
    config_not_found,
    config_helper,
    TeralityConfig,
    TeralityCredentials,
    write_output,
)
from .globals import global_client
from .data_transfers import (
    upload_local_files,
    upload_s3_files,
    copy_to_user_s3_bucket,
    DataTransfer,
)
from .encoding import encode, decode

# noinspection PyProtectedMember
from .terality_structures import (
    get_dynamic_top_level_attribute,
    top_level_functions,
    NDArray,
    Index,
    MultiIndex,
    Series,
    DataFrame,
)



from .config import TransferConfigLocal, TransferConfig, SessionInfoLocal, SessionInfo
from .structs import (
    StructRef,
    NDArrayMetadata,
    PandasIndexMetadata,
    PandasSeriesMetadata,
    PandasDFMetadata,
    Display,
    PandasFunctionRequest,
)
from .requests_responses import (
    CreateSessionResponse,
    ComputationResponse,
    DeleteSessionResponse,
    PendingComputationResponse,
    ErrorResponse,
    UploadRequest,
    ExportRequest,
    ExportResponse,
    DataTransferResponse,
    AwsCredentials,
    ImportFromCloudRequest,
    ImportFromCloudResponse,
    StorageService,
    AwsPresignedUrlSource,
    ObjectStorageKey,
    ImportFromS3Source,
    AwsS3ObjectPartExportRequest,
    AwsS3PartsExport,
    ExportToCloudRequest,
    ExportToS3Response,
)



from typing import Set

import pandas as pd
from terality_serde import IndexType

from . import ClassMethod, Struct, StructIterator


class ClassMethodIndex(ClassMethod):
    _class_name: str = IndexType.INDEX
    _pandas_class = pd.Index


class Index(Struct, metaclass=ClassMethodIndex):
    _pandas_class_instance = pd.Index([])
    _accessors = {"str"}
    _additional_methods = Struct._additional_methods | {"get_range_auto"}
    _indexers: Set[str] = set()

    def __iter__(self):
        return StructIterator(self)

    def to_list(self):
        pd_index = self._call_method(None, "to_pandas")
        return pd_index.to_list()

    def tolist(self):
        return self.to_list()


class ClassMethodMultiIndex(ClassMethodIndex):
    _class_name: str = IndexType.MULTI_INDEX
    _pandas_class = pd.MultiIndex


class MultiIndex(Index, metaclass=ClassMethodMultiIndex):
    _pandas_class_instance = pd.MultiIndex(levels=[[0, 1], [2, 3]], codes=[[0, 1], [0, 1]])


class ClassMethodDatetimeIndex(ClassMethodIndex):
    _class_name: str = IndexType.DATETIME_INDEX
    _pandas_class = pd.DatetimeIndex


class DatetimeIndex(Index, metaclass=ClassMethodDatetimeIndex):
    _pandas_class_instance = pd.DatetimeIndex([])


class ClassMethodInt64Index(ClassMethodIndex):
    _class_name: str = IndexType.INT64_INDEX
    _pandas_class = pd.Int64Index


class Int64Index(Index, metaclass=ClassMethodInt64Index):
    _pandas_class_instance = pd.Int64Index([])


class ClassMethodFloat64Index(ClassMethodIndex):
    _class_name: str = IndexType.FLOAT64_INDEX
    _pandas_class = pd.Float64Index


class Float64Index(Index, metaclass=ClassMethodFloat64Index):
    _pandas_class_instance = pd.Float64Index([])



from .pandas_job import call_pandas_function
from .top_level import get_dynamic_top_level_attribute, top_level_functions
from .structure import ClassMethod, Struct, StructIterator
from .ndarray import NDArray
from .index import Index, MultiIndex
from .series import Series
from .dataframe import DataFrame
from .group_by import DataFrameGroupBy, SeriesGroupBy



from .exceptions import TeralitySerdeError, TeralitySerializationError, TeralityDeserializationError
from .external_classes import (
    ExternalTypeSerializer,
    TypeSerde,
    TupleSerde,
    DateTimeSerde,
    scalar_types,
    all_external_types,
    external_base_classes,
)
from .serde_mixin import SerdeMixin, loads, dumps, SerializableEnum, loads_async, dumps_async
from .recursive_apply import (
    apply_func_on_object_recursively,
    apply_async_func_on_object_recursively,
)
from .callables import CallableWrapper
from .struct_types import IndexType, StructType



from itertools import groupby
from operator import attrgetter, itemgetter
from typing import Dict, Iterator, List, Tuple, Optional

from botocore.client import BaseClient
from botocore.exceptions import ClientError, NoCredentialsError
from common_client_scheduler import AwsCredentials
from common_client_scheduler.requests_responses import (
    AwsPresignedUrlSource,
    ObjectStorageKey,
    ImportFromCloudRequest,
    ImportFromS3Source,
    StorageService,
)
from common_client_scheduler.requests_responses import (
    AwsS3ObjectPartExportRequest,
    ExportToCloudRequest,
    AwsS3PartsExport,
    ExportToS3Response,
)

from .. import logger, global_client
from .common import ExportedS3File, OBJECT_CHUNK_SIZE_BYTES, list_all_objects_under_prefix, S3


def _generate_presigned_urls(
    client: BaseClient,
    s3_bucket: str,
    s3_key: str,
    object_size: int,
    part_size_bytes: int = 200 * 1024 * 1024,
) -> Iterator[AwsPresignedUrlSource]:
    """Generate pre-signed URLs allowing to read part of a S3 object.

    This function provides a few guarantees, that the server is relying on:
    * for each object, the URLs cover contiguous byte ranges of this object
    * the URLs are sorted by part number
    """
    range_start_byte = 0
    while range_start_byte < object_size:
        # byte range are inclusive
        range_end_byte = min(range_start_byte + part_size_bytes, object_size) - 1
        expiration = 3600
        url = client.generate_presigned_url(
            "get_object",
            Params={"Bucket": s3_bucket, "Key": s3_key},
            ExpiresIn=expiration,
        )
        yield AwsPresignedUrlSource(
            url=url,
            range_start_byte=range_start_byte,
            range_end_byte=range_end_byte,
            object_key=ObjectStorageKey(bucket=s3_bucket, key=s3_key),
        )
        range_start_byte = range_end_byte + 1


def _generate_upload_presigned_urls(client, s3_bucket: str, s3_key_prefix: str) -> list:
    presigned_urls = []
    for s3_key, object_size in list_all_objects_under_prefix(client, s3_bucket, s3_key_prefix):
        presigned_urls += list(_generate_presigned_urls(client, s3_bucket, s3_key, object_size))
    return presigned_urls


def upload_s3_files(s3_bucket: str, s3_key_prefix: str) -> str:
    """
    Copy files from a source (user-owned) S3 bucket to a Terality-owned S3 bucket.

    Generate pre-signed URL for each chunk of each object to transfer, then send all these URLs to the
    Terality API. Terality will perform the copy server-side. The user AWS credentials are only used to
    generate these pre-signed URLs and are never sent to Terality.

    Assumes that the user has read permissions on all objects to be copied. Uses the boto3 standard
    credentials discovery mecanism.

    Args:
        s3_bucket: source S3 bucket
        s3_key_prefix: source S3 objects

    Return:
        a transfer ID. This ID can then be used as input to "read_csv/read_parquet" requests.
    """
    try:
        presigned_urls = _generate_upload_presigned_urls(S3.client(), s3_bucket, s3_key_prefix)
    except (ClientError, NoCredentialsError) as e:
        # retry as an anonymous client if the user IAM account blocks S3 access to buckets in other accounts, or the client has no credentials
        try:
            presigned_urls = _generate_upload_presigned_urls(
                S3.client_anonymous(), s3_bucket, s3_key_prefix
            )
        except Exception:
            # if it fails too re-raise the original exception
            raise e from None
    request = ImportFromCloudRequest(
        service=StorageService.AWS_S3,
        source=ImportFromS3Source(presigned_urls=presigned_urls),
    )
    response = global_client().poll_for_answer("imports/cloud", request)
    return response.transfer_id


def copy_to_user_s3_bucket(
    aws_credentials: AwsCredentials,
    transfer_id: str,
    aws_region: Optional[str],
    destination_s3_path: str,
) -> None:
    """Copy the files identified by `transfer_id` from a Terality export bucket to a user S3 bucket.

    Args:
        aws_credentials: AWS credentials to use to read files in the Terality export bucket
        transfer_id: transfer ID used to identify the files to copy from the Terality export bucket
        aws_region: AWS region hosting the Terality export bucket
        destination_s3_path: a string of the form 's3://bucket/some/key/*.extension'. The '*' will be replaced
            by integers (starting from 0) in the destination.
    """
    if not destination_s3_path.startswith("s3://"):
        raise ValueError(
            f"destination_s3_path must start with 's3://', got '{destination_s3_path}'"
        )

    download_config = global_client().get_download_config()
    source_bucket = download_config.bucket_region(aws_region)
    source_prefix = f"{download_config.key_prefix}{transfer_id}/"

    destination_s3_bucket, destination_s3_prefix_template = destination_s3_path[
        len("s3://") :
    ].split("/", maxsplit=1)

    auth_s3_client = S3.client_from_credentials(aws_credentials)
    files = [
        (
            destination_s3_prefix_template.replace("*", str(i)),
            ExportedS3File(bucket=source_bucket, key=s3_key, size_bytes=object_size),
        )
        for i, (s3_key, object_size) in enumerate(
            list_all_objects_under_prefix(auth_s3_client, source_bucket, source_prefix)
        )
    ]

    s3_client = S3.client()
    _copy_exported_files_to_s3_destination(s3_client, files, destination_s3_bucket)


def _copy_exported_files_to_s3_destination(  # pylint: disable=too-many-locals
    s3_client, files_to_copy: List[Tuple[str, ExportedS3File]], destination_s3_bucket: str
) -> None:
    """Copy S3 objects to a destination bucket.

    Args:
        s3_client: a botocore S3 client with enough permissions to write objects to the destination bucket
        files_to_copy: a list of (destination S3 key, source file to copy)
        destination_s3_bucket: destination S3 bucket
    """
    # TODO: this method is too long, clean that up
    presigned_urls = []
    uploads_in_progress: Dict[str, Tuple[str, str]] = {}  # dict of upload_id => (bucket, key)
    try:
        # Start multipart uploads and prepare presigned URLs
        for destination_s3_key, file_to_copy in files_to_copy:
            res = s3_client.create_multipart_upload(
                Bucket=destination_s3_bucket,
                Key=destination_s3_key,
            )
            upload_id = res["UploadId"]
            uploads_in_progress[upload_id] = (destination_s3_bucket, destination_s3_key)
            urls = _generate_upload_part_presigned_urls(
                s3_client,
                upload_id=upload_id,
                destination_s3_bucket=destination_s3_bucket,
                destination_s3_key=destination_s3_key,
                object_size_bytes=file_to_copy.size_bytes,
            )
            presigned_urls += [
                AwsS3ObjectPartExportRequest(
                    presigned_url=url[0],
                    destination_object_key=ObjectStorageKey(
                        bucket=destination_s3_bucket, key=destination_s3_key
                    ),
                    source_object_key=ObjectStorageKey(
                        bucket=file_to_copy.bucket, key=file_to_copy.key
                    ),
                    multipart_upload_id=upload_id,
                    part_number=part_number,
                    range_start_byte=url[1],
                    range_end_byte=url[2],
                )
                for part_number, url in enumerate(
                    urls, 1
                )  # part numbers start at 1 in the AWS S3 API
            ]

        # Send request to the API
        request = ExportToCloudRequest(
            service=StorageService.AWS_S3,
            export_request=AwsS3PartsExport(parts=presigned_urls),
        )
        export_res: ExportToS3Response = global_client().poll_for_answer("exports/cloud", request)

        # Complete the uploads
        parts = export_res.uploaded_parts
        keyfunc = attrgetter("multipart_upload_id")
        parts_by_upload_id = groupby(sorted(parts, key=keyfunc), key=keyfunc)
        for upload_id, parts in parts_by_upload_id:  # type: ignore
            parts = list(parts)
            etags = [
                {
                    "PartNumber": part.part_number,
                    "ETag": part.etag,
                }
                for part in parts
            ]
            bucket = parts[0].destination_object_key.bucket
            key = parts[0].destination_object_key.key
            etags = sorted(etags, key=itemgetter("PartNumber"))
            s3_client.complete_multipart_upload(
                UploadId=upload_id,
                Bucket=bucket,
                Key=key,
                MultipartUpload={"Parts": etags},
            )
            del uploads_in_progress[upload_id]
    finally:
        # Cancel uploads that were not completed to avoid the user being charged for data
        for upload_id, (bucket, key) in uploads_in_progress.items():
            try:
                logger.info(
                    f"Aborting multipart upload on s3://{bucket}/{key} (upload ID {upload_id})."
                )
                s3_client.abort_multipart_upload(Bucket=bucket, Key=key, UploadId=upload_id)
            except Exception:
                logger.exception(
                    f"Error when aborting multipart upload on s3://{bucket}/{key} (upload ID {upload_id})."
                )


def _generate_upload_part_presigned_urls(
    s3_client: BaseClient,
    *,
    upload_id: str,
    destination_s3_bucket: str,
    destination_s3_key: str,
    object_size_bytes: int,
    part_size_bytes: int = OBJECT_CHUNK_SIZE_BYTES,
) -> List[Tuple[str, int, int]]:
    """
    Generate presigned URLs for the AWS UploadPart S3 API.

    To write an object of size `object_size_bytes`, we'll call UploadPart on each object part. Thus,
    we need to generate one presigned URL per part.

    Args:
        s3_client: a S3 boto3 client
        upload_id: multipart upload ID, as returned by CreateMultipartUpload
        destination_s3_bucket: S3 bucket containing the destination S3 object
        destination_s3_key: S3 key of the object to write
        object_size_bytes: total size, in bytes, of the object to write
        part_size_bytes: maximum size of a part to upload.

    Return:
        a list of tuples (presigned URL, start byte, end byte)
    """
    expiration = 3600
    range_start_byte = 0
    part_number = 1
    out = []
    while range_start_byte < object_size_bytes:
        range_end_byte = min(range_start_byte + part_size_bytes, object_size_bytes) - 1
        url = s3_client.generate_presigned_url(
            ClientMethod="upload_part",
            Params={
                "Bucket": destination_s3_bucket,
                "Key": destination_s3_key,
                "UploadId": upload_id,
                "PartNumber": part_number,
            },
            ExpiresIn=expiration,
        )
        out.append((url, range_start_byte, range_end_byte))

        part_number += 1
        range_start_byte = range_end_byte + 1
    return out



"""Implement utilities for transferring data to/from Terality.

Data transfers are usually done in two separate steps, one transfer between Terality internal systems
and a Terality export S3 bucket, and one transfer between a Terality export S3 bucket and a user-controlled
system. This two-step process has the advantage of simplicity (the implementation of these two steps
is decoupled).

When copying between Terality transfer buckets and other cloud storage systems (AWS S3, Azure Datalake...),
the copy is performed server-side to avoid the client having to download and reupload files.

There are some details in the implementation that are mostly the result of incremental design decisions.
Some simplifications are probably possible. Below is a high level overview of the flow.

Security
~~~~~~~~

At no point the Terality API gets access to the user cloud credentials. When copying to or from a cloud
storage provider, the Terality client generates pre-signed URLs (or the equivalent for the cloud provider
at hand) using the user local credentials, and send these URLs to the server.  These URLs have a limited
expiry time and don't carry any privileges (outside the pre-signed API call).

When the client calls the S3 API on a Terality export bucket, it uses temporary credentials generated
by the Terality server. We don't depend on the user credentials having enough S3 permissions to access the
Terality bucket.

Imports
~~~~~~~

First step: copy from the user-controlled system to the Terality export S3 bucket
 -> For local files, directly upload the file using the boto3 library (using credentials provided by
    the Terality API).
 -> For files stored in a S3 bucket, generate pre-signed URLs for the GetObject operation (using user
    credentials as discovered by boto3), send these URLs to the Terality API. The copy is implemented
    server-side.
 -> For files stored in an Azure Gen2 Datalake filesystem, generate shared access signature for each
    file (= blob) to copy. Send these signatures to the Terality API. The copy is implemented server-side.
This generates a transfer ID.

Second step: import data into Terality
 -> Call the `read_csv`, `read_parquet`... pandas function passing the location (transfer ID + AWS region)
    of the S3 objects in the Terality import bucket.

Exports
~~~~~~~

First step: export data from Terality to a Terality export bucket
 -> Call the `to_csv`, `to_parquet`... pandas function. The API returns a transfer ID.

Second step: copy from the Terality export S3 bucket to the user-controlled system
 -> Download to a local file: directly download the file using the boto3 library
    (using credentials provided by the Terality API).
 -> Copy to a S3 bucket: the client starts a multipart upload (using user credentials as discovered by
    boto3), generates a list of presigned URLs (one for each part to write), send these URLs to the server,
    then complete the multipart upload. The actual copy is performed server-side.
 -> Copy to an Azure Gen2 Datalake filesystem: the client generates shared access signature for each
    file (= blob) to copy. The client sends these signatures to the Terality API. The copy is
    implemented server-side.
"""
from .common import (
    DataTransfer,
)
from .s3 import (
    upload_s3_files,
    copy_to_user_s3_bucket,
)
from .local import (
    upload_local_files,
)

# Don't import Azure here, because the Azure dependency is optional and the import may fail.
